package com.liangjian.web.common.service.crud;

import com.liangjian.web.common.dynamic.DynamicFilter;
import com.liangjian.web.common.dynamic.DynamicParams;
import com.liangjian.web.common.dynamic.SortDirection;
import com.liangjian.web.common.entity.helper.PageWrapper;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.data.jpa.repository.JpaSpecificationExecutor;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;
import java.util.List;

/**
 * Created by huangziqi on 2020/1/13
 */
public abstract class JpaDynamicQueryService<T> implements DynamicApiService<T> {

    /**
     * 获取Jpa的复杂查询构造器
     * @return
     */
    public abstract JpaSpecificationExecutor<T> getJpaSpecificationExecutor();

    /**
     * 转换DynamicFilter成Jpa的条件构造器Predicate
     * @param filter
     * @param root
     * @param criteriaBuilder
     * @param <A>
     * @return
     */
    protected  <A> Predicate toJpaSpecification(DynamicFilter filter, Root<A> root, CriteriaBuilder criteriaBuilder) {
        switch (filter.getOperate()) {
            case Equal: {
                Path path = root.get(filter.getField());
                return criteriaBuilder.equal(path, filter.transportType(path.getJavaType()));
            }
            case NotEqual:{
                Path path = root.get(filter.getField());
                return criteriaBuilder.notEqual(path, filter.transportType(path.getJavaType()));
            }
            case Less: return criteriaBuilder.lt(root.get(filter.getField()).as(Double.class), Double.valueOf(filter.getValue()));
            case LessEqual: return criteriaBuilder.lt(root.get(filter.getField()).as(Double.class), Double.valueOf(filter.getValue()));
            case Greater: return criteriaBuilder.gt(root.get(filter.getField()).as(Double.class), Double.valueOf(filter.getValue()));
            case GreaterEqual: return criteriaBuilder.gt(root.get(filter.getField()).as(Double.class), Double.valueOf(filter.getValue()));
            case Contains: return criteriaBuilder.like(root.get(filter.getField()), "%"+filter.getValue()+"%");
            case StartWith: return criteriaBuilder.like(root.get(filter.getField()), filter.getValue()+"%");
            case EndWith: return criteriaBuilder.like(root.get(filter.getField()), "%"+filter.getValue());
            default: throw new RuntimeException("未知对应判断操作符"+filter.getOperate());
        }
    }

    /**
     * 转换SortDirection成Jpa的Sort
     * @param sortDirection
     * @return
     */
    protected Sort.Direction toJpaPageSort(SortDirection sortDirection) {
        if(sortDirection == SortDirection.ASC) {
            return Sort.Direction.ASC;
        } else {
            return Sort.Direction.DESC;
        }
    }

    /**
     * 转换DynamicParams成Jpa复杂条件
     * @param params
     * @return
     */
    protected Specification<T> getDynamicSpecification(DynamicParams params) {
        Specification<T> specification = (root, criteriaQuery, criteriaBuilder) ->
                criteriaBuilder.and(params.getFilters()
                        .stream()
                        .map(i -> toJpaSpecification(i,root,criteriaBuilder))
                        .toArray(Predicate[]::new));
        return specification;
    }

    @Override
    public PageWrapper<T> getPageByCondition(DynamicParams params) {
        Pageable pageable = null;
        if(params.getSortField() != null && !params.getSortField().isEmpty()) {
            SortDirection sortDirection = params.getSortDirection() == null ? SortDirection.ASC : params.getSortDirection();
            pageable = PageRequest.of(params.getPage() - 1, params.getSize(), toJpaPageSort(sortDirection), params.getSortField());
        } else {
            pageable = PageRequest.of(params.getPage() - 1, params.getSize());
        }
        Specification<T> specification = getDynamicSpecification(params);
        Page<T> paged = getJpaSpecificationExecutor().findAll(specification, pageable);
        return PageWrapper.wrap(paged);
    }

    @Override
    public List<T> getListByCondition(DynamicParams params) {
        List<T> list = null;
        Specification<T> specification = getDynamicSpecification(params);
        if(params.getSortField() != null && !params.getSortField().isEmpty()) {
            SortDirection sortDirection = params.getSortDirection() == null ? SortDirection.ASC : params.getSortDirection();
            list = getJpaSpecificationExecutor().findAll(specification, Sort.by(toJpaPageSort(sortDirection),params.getSortField()));
        } else {
            list = getJpaSpecificationExecutor().findAll(specification);
        }
        return list;
    }
}
